from rdflib import Graph, URIRef
from rdflib.term import Node
from rdflib.namespace import Namespace
from StringIO import StringIO
import networkx as nx
from shortuuid import uuid

SCM = Namespace('http://www.smartdeveloperhub.org/vocabulary/scm#')

g = Graph()

with open('fragment.ttl') as ttl:
    g.parse(StringIO(ttl.read()), format='n3')


def printable(x):
    if isinstance(x, Node):
        return x.toPython()
    return x


def tp_parts(tp):
    """
    :param tp: A triple pattern string
    :return: A string-based 3-tuple like (subject, predicate, object)
    """
    if tp.endswith('"'):
        parts = [tp[tp.find('"'):]]
        st = tp.replace(parts[0], '').rstrip()
        parts = st.split(" ") + parts
    else:
        parts = tp.split(' ')
    return tuple(parts)


def __transform(x):
    """
    Trick to avoid literal language tags problem, etc.
    """
    if x.startswith('"'):
        var = uuid()
        return '?%s FILTER(str(?%s) = %s)' % (var, var, x)
    return x


def build_query(gp, projection):
    def __build_query_from(x, depth=0):
        def build_pattern_query((u, v, data)):
            return '\n%sOPTIONAL { %s %s %s %s }' % ('\t' * (depth + 1), u, data['predicate'], v,__build_query_from(v, depth + 1))
            # if gp_graph.out_degree(v) == 0:
            #     return '\n%sOPTIONAL { %s %s %s}' % ('\t' * (depth + 1), u, data['predicate'], v)
            # else:
            #     x_query = '\n%s%s %s %s %s .' % (
            #         '\t' * (depth + 1), u, data['predicate'], v, __build_query_from(v, depth + 1))
            #     x_query = '\n%sOPTIONAL { %s \n%s}' % ('\t' * (depth + 1), x_query, '\t' * depth)
            #     return x_query

        print depth
        out_edges = list(gp_graph.out_edges_iter(x, data=True))
        out_edges = reversed(sorted(out_edges, key=lambda x: gp_graph.out_degree))
        if out_edges:
            return ' '.join([build_pattern_query(x) for x in out_edges])
        return ''

    gp = filter(lambda x: ' a ' not in x and 'rdf:type' not in x, gp)
    gp_parts = [[__transform(part) for part in tp_parts(tp)] for tp in gp]

    blocks = []
    gp_graph = nx.DiGraph()
    filter_block = []
    for gp_part in gp_parts:
        if 'FILTER' not in gp_part[2]:
            gp_graph.add_edge(gp_part[0], gp_part[2], predicate=gp_part[1])
        else:
            filter_block.append(' '.join(gp_part))

    roots = filter(lambda x: gp_graph.in_degree(x) == 0, gp_graph.nodes())

    blocks += [' %s a agora:Root \n OPTIONAL { %s } .' % (root, __build_query_from(root)) for root in roots]
    if filter_block:
        blocks.append('{ %s }' % ' .\n '.join(filter_block))

    where_gp = ' .\n'.join(blocks)
    query = """SELECT DISTINCT %s WHERE {\n%s\n}""" % (projection, where_gp)
    return query


# qr = g.query("""SELECT ?rn ?bn ?cid ?c ?s ?b WHERE {
#                     {
#                         OPTIONAL { ?s doap:name ?rn }
#                         OPTIONAL { ?s scm:hasBranch ?b
#                             OPTIONAL { ?b doap:name ?bn }
#                             OPTIONAL { ?b scm:hasCommit ?c
#                                 OPTIONAL { ?c scm:commitId ?cid }
#                             }
#                         }
#                     }
#                 }""")

# query = """SELECT DISTINCT ?rn ?bn ?s ?b WHERE {
#                     OPTIONAL {
#                         ?s scm:hasBranch ?b
#                             OPTIONAL {
#                                 ?b doap:name ?bn
#                             }
#                             OPTIONAL {
#                                 ?b scm:hasCommit ?c
#                                 OPTIONAL {
#                                     ?c scm:commitId ?cid
#                                 }
#                             }
#                         OPTIONAL { ?s doap:name ?rn }
#                     }
#                 }"""

# query = """SELECT DISTINCT ?n ?fc ?nk ?d ?r ?b WHERE {
# 	?r a agora:Root
# 	OPTIONAL {
# 		OPTIONAL { ?r scm:hasBranch ?b }
# 		OPTIONAL { ?r doap:developer ?d OPTIONAL { ?d scm:firstCommit ?fc } OPTIONAL { ?d foaf:nick ?nk } }
# 		OPTIONAL { ?r doap:name ?n }
# 	} .
# }"""

query = build_query(
    [
        '?r doap:name ?rn',
        '?r doap:developer ?d',
        '?d foaf:nick ?nk',
        '?d scm:firstCommit ?fc',
        '?r scm:hasBranch ?b'
    ], '?r ?d ?b ?rn ?nk ?fc')
print query

qr = g.query(query)
for r in qr:
    print map(printable, r)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
from datetime import datetime as time

from agora.stoa.client import get_fragment_generator, get_enrichment_generator, get_query_generator
import logging

__author__ = 'Fernando Serena'

ch = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
ch.setLevel(logging.DEBUG)
logger = logging.getLogger('agora')
logger.addHandler(ch)
logger.setLevel(logging.DEBUG)

AGORA_HOST = '138.4.249.224'
AGORA_PORT = 9009

# BROKER_HOST = '138.4.249.161'
# AGORA_HOST = '138.4.249.161'
# AGORA_PORT = 9002

BROKER_HOST = '192.168.1.43'
# AGORA_HOST = '192.168.1.43'
# BROKER_HOST = '138.4.220.50'
# AGORA_HOST = '138.4.220.50'

# prefixes, gen = get_enrichment_generator('?s a dbpedia-owl:Film',
#                                          '?s foaf:name "Alice"',
#                                          '* amovies:hasFilm ?s',
#                                          broker_host=BROKER_HOST,
#                                          agora_host=AGORA_HOST,
#                                          agora_port=9002,
#                                          target='http://localhost:5005/',
#                                          wait=True,
#                                          exchange='sdh',
#                                          topic_pattern='curator.request')

# prefixes, gen = get_enrichment_generator('?s dbpedia-owl:starring ?a',
#                                          '* amovies:hasActor ?a',
#                                          broker_host=BROKER_HOST,
#                                          agora_host=AGORA_HOST,
#                                          agora_port=9002,
#                                          target='http://localhost:5005/',
#                                          wait=True,
#                                          exchange='sdh',
#                                          topic_pattern='curator.request')

# prefixes, gen = get_enrichment_generator('?s a dbpedia-owl:Film',
#                                          '?s foaf:name "2046"',
#                                          '* amovies:hasFilm ?s',
#                                          broker_host=BROKER_HOST,
#                                          agora_host=AGORA_HOST,
#                                          agora_port=AGORA_PORT,
#                                          target='http://localhost:5005/',
#                                          wait=True,
#                                          exchange='sdh',
#                                          topic_pattern='curator.request',
#                                          response_prefix='curator.response')

# prefixes, gen = get_enrichment_generator('?s dbpedia-owl:starring ?a',
#                                          '?a dbp:birthName "Paul Frederic Simon"',
#                                          '* amovies:hasActor ?a',
#                                          broker_host=BROKER_HOST,
#                                          agora_host=AGORA_HOST,
#                                          agora_port=AGORA_PORT,
#                                          target='http://localhost:5005/',
#                                          wait=True,
#                                          exchange='sdh',
#                                          topic_pattern='curator.request',
#                                          response_prefix='curator.response')


# prefixes, gen = get_fragment_generator('?s dbpedia-owl:starring ?a',
#                                        '?s foaf:name "2046"',
#                                        broker_host='138.4.249.164',
#                                        agora_host='138.4.249.164',
#                                        agora_port=9002, wait=True)

# prefixes, gen = get_query_generator('?_s foaf:name "20,000 Leagues Under the Sea"',
#                                        '?_s dbpedia-owl:starring ?a',
#                                        '?a foaf:depiction ?im',
#                                        '?a dbp:birthName ?bn',
#                                        broker_host='192.168.1.43',
#                                        agora_host='192.168.1.43',
#                                        agora_port=9002, wait=True)

# prefixes, gen = get_fragment_generator('<http://dbpedia.org/resource/55_Days_at_Peking> dbpedia-owl:starring ?a',
#                                        '<http://dbpedia.org/resource/55_Days_at_Peking> foaf:name ?n',
#                                        broker_host=BROKER_HOST,
#                                        agora_host=AGORA_HOST,
#                                        agora_port=AGORA_PORT, wait=True)

# prefixes, gen = get_query_generator('?s dbpedia-owl:starring ?a',
#                                        '?a dbp:birthName ?n',
#                                        broker_host=BROKER_HOST,
#                                        agora_host=AGORA_HOST,
#                                        agora_port=AGORA_PORT, wait=True)

# prefixes, gen = get_query_generator('<http://dbpedia.org/resource/55_Days_at_Peking> dbpedia-owl:starring ?a',
#                                        '<http://dbpedia.org/resource/55_Days_at_Peking> foaf:name ?n',
#                                        '?h amovies:hasFilm <http://dbpedia.org/resource/55_Days_at_Peking>',
#                                        broker_host='192.168.1.43',
#                                        agora_host='192.168.1.43',
#                                        agora_port=9002, wait=True)


# prefixes, gen = get_fragment_generator('?s doap:name ?rn', '?s scm:hasBranch ?b', '?b doap:name ?bn', '?b scm:hasCommit ?c',
#                                     '?c scm:commitId ?cid',
#                                     broker_host=BROKER_HOST,
#                                     agora_host=AGORA_HOST,
#                                     agora_port=AGORA_PORT, wait=True)

prefixes, gen = get_enrichment_generator('?r a scm:Repository',
                                         '?r doap:name "sdh-ci-metrics"',
                                         '?r scm:hasBranch ?b',
                                         '?b doap:name "v2"',
                                         '?b scm:hasCommit ?c',
                                         '?c scm:commitId "513e8d4346ecfd83b5e8ad16d812c9d87bbcf38e"',
                                         '* ci:forCommit ?c',
                                         '* ci:forBranch ?b',
                                         broker_host=BROKER_HOST,
                                         agora_host=AGORA_HOST,
                                         agora_port=AGORA_PORT,
                                         target='http://localhost:5005/',
                                         wait=True,
                                         exchange='sdh',
                                         topic_pattern='curator.request',
                                         response_prefix='curator.response')

# prefixes, gen = get_enrichment_generator('?r a scm:Repository',
#                                          '?r doap:name "agora-fountain"',
#                                          '?r scm:hasBranch ?b',
#                                          '?b doap:name "develop"',
#                                          '* scm:forBranch ?b',
#                                          broker_host=BROKER_HOST,
#                                          agora_host=AGORA_HOST,
#                                          agora_port=AGORA_PORT,
#                                          target='http://localhost:5005/',
#                                          wait=True,
#                                          exchange='sdh',
#                                          topic_pattern='curator.request',
#                                          response_prefix='curator.response')

pre = time.now()
for headers, content in gen:
    format = headers.get('format')
    if format == 'turtle':
        print content.serialize(format='turtle')
    else:
        print content
post = time.now()
print 'Elapsed time: {}'.format((post - pre).total_seconds())
